#ifndef PCFS_H
#define PCFS_H

/*
 *
 *  Adapted for System V Release 4	(ESIX 4.0.4)	
 *
 *  Gerard van Dorth	(gdorth@nl.oracle.com)
 *  Paul Bauwens	(paul@pphbau.atr.bso.nl)
 *
 *  May 1993
 *
 *  Originally written by Paul Popelka (paulp@uts.amdahl.com)
 *
 *  You can do anything you want with this software,
 *    just don't say you wrote it,
 *    and don't remove this notice.
 *
 *  This software is provided "as is".
 *
 *  The author supplies this software to be publicly
 *  redistributed on the understanding that the author
 *  is not responsible for the correct functioning of
 *  this software in any circumstances and is not liable
 *  for any damages caused by this software.
 *
 *  October 1992
 *
 */

#ident "@(#)pcfs.h	1.2 93/07/17 "

#include "sys/types.h"
#include "sys/param.h"
#include "sys/fcntl.h"
#include "sys/proc.h"
#include "sys/user.h"
#include "sys/stat.h"
#include "sys/statvfs.h"
#include "sys/dirent.h"
#include "sys/sysmacros.h"
#include "sys/vfs.h"
#include "sys/vnode.h"
#include "sys/buf.h"
#include "sys/cmn_err.h"
#include "sys/conf.h"
#include "sys/cred.h"
#include "sys/ddi.h"
#include "sys/debug.h"
#include "sys/disp.h"
#include "sys/dnlc.h"
#include "sys/errno.h"
#include "sys/fbuf.h"
#include "sys/file.h"
#include "sys/flock.h"
#include "sys/immu.h"
#include "sys/inode.h"
#include "sys/kmem.h"
#include "sys/mkdev.h"
#include "sys/mman.h"
#include "sys/mount.h"
#include "sys/open.h"
#include "sys/pathname.h"
#include "sys/resource.h"
#include "sys/swap.h"
#include "sys/sysinfo.h"
#include "sys/systm.h"
#include "sys/time.h"
#include "sys/uio.h"

#include "fs/fs_subr.h"

/* pcfs includes */
#include "bootsect.h"
#include "bpb.h"
#include "direntry.h"
#include "denode.h"
#include "fat.h"
#include "pcfs_filsys.h"
#include "pcfs_lbuf.h"

#ifdef _KERNEL

/* pcfs prototypes */

int createde(struct denode *dep, struct vnode *dvp, struct denode **depp);
int deflush(struct vfs *vfsp, int force);
int detrunc(struct denode *dep, u_long length, int flags);
int deupdat(struct denode *dep, timestruc_t *tp, int waitfor);
int doscheckpath(struct denode *source, struct denode *target);
int dosdirempty(struct denode *dep);
int extendfile(struct denode *dep, lbuf_t **bpp, u_long *ncp);
int fillinusemap(struct pcfs_vfs *pvp);
int get_direntp(struct pcfs_vfs *pvp, u_long dirclust, u_long diroffset, dosdirent_t *dep);
int readde(struct denode *dep, lbuf_t **bpp, struct direntry **epp);
int removede(struct vnode *vp);
int strcpy(char *, const char *);
int strlen(const char *);
void cleanlocks();
void fc_purge(struct denode *dep, unsigned int frcn);
void fc_lookup(struct denode *dep, u_long findcn, u_long *frcnp, u_long *fsrcnp);
void delock (struct denode *dep);
void deput(struct denode *dep);
void dereject (denode_t *dp);
void deunhash(struct denode *dp);
void deunlock (struct denode *dep);
void pcfs_deflush(void);
void reinsert(struct denode *dep);

extern struct vnodeops pcfs_vnodeops;

#ifndef NO_GENOFF
extern int pcfs_shared;
#endif

#endif /* _KERNEL */

#endif /* PCFS_H */
